import math
import numpy as np
import argparse
import tqdm
import torch
import torch.nn as nn
import torch.nn.functional as F
from pathlib import Path
from collections import defaultdict

import zouqi
from torchzq.runners.base import BaseRunner
from torchzq.scheduler import Scheduler


class CombinedOptimizer(list):
    def __init__(self, optimizer):
        super().__init__(optimizer)

    def state_dict(self):
        state_dicts = []
        for optimizer in self:
            state_dicts.append(optimizer.state_dict())
        return state_dicts

    def load_state_dict(self, state_dicts):
        for i, optimizer in enumerate(self):
            optimizer.load_state_dict(state_dicts[i])


class GANRunner(BaseRunner):
    def __init__(
        self,
        *args,
        gp_weight: float = 10,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.update_args(dict(gp_weight=gp_weight))

    def gp_loss(self, images, outputs):
        args = self.args

        if args.gp_weight <= 0:
            return 0

        n = images.shape[0]

        try:
            grad = torch.autograd.grad(
                outputs=outputs,
                inputs=images,
                grad_outputs=torch.ones_like(outputs),
                create_graph=True,
                retain_graph=True,
                only_inputs=True,
            )[0]
        except:
            # during validation, no_grad is set.
            return torch.zeros([])

        grad = grad.reshape(n, -1)

        return args.gp_weight * ((grad.norm(2, dim=1) - 1) ** 2).mean()

    def create_optimizer(self, model):
        return CombinedOptimizer(
            [
                super().create_optimizer(model[0]),
                super().create_optimizer(model[1]),
            ]
        )

    def create_scheduler(self):
        args = self.args
        scheduler = Scheduler()
        if self.training:
            args.g_lr = scheduler.schedule(args.g_lr)
            args.d_lr = scheduler.schedule(args.d_lr)
        return scheduler

    @staticmethod
    def feed(model, z, label):
        try:
            return model(z, label)
        except TypeError:
            return model(z)

    def step(self, batch):
        args = self.args

        stats = {}

        logger = self.logger

        real, label = self.prepare_batch(batch)

        G, D = self.model

        if self.training:
            g_optimizer, d_optimizer = self.optimizer
            g_optimizer.set_lr(args.g_lr())
            d_optimizer.set_lr(args.d_lr())

        # train d
        if self.training:
            d_optimizer.zero_grad()

        z = self.sample(len(real))

        with self.autocast_if_use_fp16():
            fake = self.feed(G, z, label)
            real.requires_grad_()
            fake_output = self.feed(D, fake, label)
            real_output = self.feed(D, real, label)

            losses = {
                "d_fake_loss": F.relu(1 - fake_output).mean(),
                "d_real_loss": F.relu(1 + real_output).mean(),
                "d_gp_loss": self.gp_loss(real, real_output),
            }

            d_loss = 0
            for name, loss in losses.items():
                d_loss += loss
                stats[name] = loss.item()
            stats["d_loss"] = d_loss.item()

        if self.training:
            if args.use_fp16:
                self.scaler.scale(d_loss).backward()
                self.scaler.unscale_(d_optimizer)
            else:
                d_loss.backward()

            # grad clip here (TODO)

            if args.use_fp16:
                self.scaler.step(d_optimizer)
                self.scaler.update()
            else:
                d_optimizer.step()

            stats["d_lr"] = d_optimizer.get_lr()

        # train g
        if self.training:
            g_optimizer.zero_grad()

        z = self.sample(len(real))
        with self.autocast_if_use_fp16():
            fake = self.feed(G, z, label)
            fake_output = self.feed(D, fake, label)
            g_loss = fake_output.mean()
            stats["g_loss"] = g_loss.item()

        if self.training:
            if args.use_fp16:
                self.scaler.scale(g_loss).backward()
                self.scaler.unscale_(g_optimizer)
            else:
                g_loss.backward()

            # grad clip here (TODO)

            if args.use_fp16:
                self.scaler.step(g_optimizer)
                self.scaler.update()
            else:
                g_optimizer.step()

            stats["g_lr"] = g_optimizer.get_lr()

        return stats

    @zouqi.command(inherit=True)
    def train(self, *args, g_lr=1e-3, d_lr=1e-3, lr=None, **kwargs):
        self.args.g_lr = g_lr
        self.args.d_lr = d_lr
        super().train(*args, **kwargs)



import pickle
import inspect
from pathlib import Path
from urllib.parse import urlencode
from collections import defaultdict


def dumbrw(*args, **kwargs):
    pass


def pickle_dumper(obj, path):
    with open(path, "wb") as f:
        return pickle.dump(obj, f)


def pickle_loader(path):
    with open(path, "rb") as f:
        return pickle.load(f)


class Cache:
    def __init__(self, folder, verbose=False):
        self.folder = Path(folder)
        self.buffer = defaultdict(dict)
        self.verbose = verbose

    def load(self, loader, name, key):
        obj = None
        if name in self.buffer and key in self.buffer[name]:
            obj = self.buffer[name][key]
        else:
            path = Path(self.folder, name, key)
            if path.exists():
                obj = loader(path)
            self.buffer[name][key] = obj
        if self.verbose and obj is not None:
            print(name, key, "hit.")
        return obj

    def dump(self, dumper, name, key, obj):
        self.buffer[name][key] = obj
        path = Path(self.folder, name, key)
        path.parent.mkdir(parents=True, exist_ok=True)
        dumper(obj, path)
        if self.verbose:
            print(name, key, "cached.")

    def __call__(self, key=urlencode, dumper=pickle_dumper, loader=pickle_loader):
        if dumper is None:
            dumper = dumbrw

        if loader is None:
            loader = dumbrw

        def wrapper(func):
            def wrapped(*args, **kwargs):
                argnames = inspect.getargspec(func)[0][: len(args)]
                kwargs = {**kwargs, **dict(zip(argnames, args))}
                s = key(kwargs)
                obj = self.load(loader, func.__name__, s)
                if obj is None:
                    obj = func(**kwargs)
                    self.dump(dumper, func.__name__, s, obj)
                return obj

            return wrapped

        return wrapper



import argparse
from collections import defaultdict


class Event(list):
    def __call__(self, *args, **kwargs):
        for f in self:
            f(*args, **kwargs)


class Events(argparse.Namespace):
    def __init__(self, *names):
        super().__init__(**{name: Event() for name in names})



import torch.nn as nn
import zouqi

from ..parsing import optional
from .base import BaseRunner


class LegacyRunner(BaseRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compute_loss(self, batch) -> dict:
        raise NotImplementedError

    def prepare_batch(self, batch):
        raise NotImplementedError

    def step(self, batch):
        args = self.args

        model = self.model
        stats = {}

        batch = self.prepare_batch(batch)

        with self.autocast_if_use_fp16():
            losses = self.compute_loss(batch)

        loss = 0
        for key, value in losses.items():
            if not value.requires_grad:
                key += "#"  # indicate no grad
            stats[key] = value.item()
            loss += value

        if self.training:
            optimizer = self.optimizer
            optimizer.set_lr(self.args.lr())

            if args.use_fp16:
                self.scaler.scale(loss / args.update_every).backward()
            else:
                (loss / args.update_every).backward()

            if model.iteration % args.update_every == 0:
                if args.use_fp16:
                    self.scaler.unscale_(optimizer)

                self.update_grad(stats)

                if args.use_fp16:
                    self.scaler.step(optimizer)
                    self.scaler.update()
                else:
                    optimizer.step()

                optimizer.zero_grad()

            stats["lr"] = optimizer.get_lr()

        return stats

    def update_grad(self, stats):
        args = self.args
        grad_norm = nn.utils.clip_grad_norm_(
            self.model.parameters(),
            args.grad_clip_thres or 1e9,
        )
        stats["grad_norm"] = grad_norm.item()

    @zouqi.command
    def train(
        self,
        update_every: int = 1,
        grad_clip_thres: optional(float) = None,
        **kwargs,
    ):
        self.update_args(locals(), ["args", "kwargs", "self"])
        super().train(**kwargs)



import re
import yaml
import argparse
from pathlib import Path


def empty_value():
    pass


def load_yaml(path, command):
    data = {}
    with open(path, "r") as f:
        # remove the blank at the end of lines
        # to ensure yaml print multilines correctly
        lines = f.read().splitlines()
        data = "\n".join([l.rstrip() for l in lines])
        data = yaml.load(data, Loader=yaml.FullLoader)
    data = data or {}

    if command in data:
        data = {**data, **data[command]}

    # except for command, only support first level
    for key in list(data.keys()):
        if type(data[key]) is dict:
            del data[key]

    return data


def update_tree(a, b):
    for k in b:
        if k in a and type(a[k]) is dict and type(b[k]) is dict:
            update_tree(a[k], b[k])
        else:
            a[k] = b[k]


def parse_yaml(path, command):
    data = load_yaml(path, command)
    result = {}
    defaults = data.get("default", [])
    if type(defaults) is not list:
        defaults = [defaults]
    for path in defaults:
        update_tree(result, parse_yaml(path, command))
    update_tree(result, data)
    return result


class ConfigParser:
    def __init__(self, path):
        self.path = Path(path)

    def parse_manual_options(self, manual_options):
        parsed = []
        for s in manual_options:
            if s.startswith("--"):
                parsed.append([s.lstrip("-")])
            else:
                parsed[-1].append(s)
        for i, s in enumerate(parsed):
            if len(s) > 2:
                parsed[i] = [s[0], s[1:]]
        return parsed

    @staticmethod
    def parse_key(key):
        return "--" + key.replace("_", "-").lstrip("-")

    @staticmethod
    def parse_value(value):
        if type(value) is list:
            return " ".join(map(ConfigParser.parse_value, value))
        elif " " in str(value) or "(" in str(value):
            value = f'"{value}"'
        return str(value)

    def parse_option(self, k, v=empty_value):
        if k == "name":
            groups = re.findall("\$from\((.+?)\)", v)
            if groups:
                path = Path(groups[0])
                if path not in self.path.parents:
                    raise ValueError(
                        "Name root should be the parent of the configuration file!"
                    )
                rpath = self.path.relative_to(path)
                v = rpath.with_suffix("")
        return f"{self.parse_key(k)} {'' if v is empty_value else self.parse_value(v)}"

    def parse_cmd(self, command, manual_arguments, manual_options):
        data = parse_yaml(self.path, command)

        if "default" in data:
            del data["default"]

        if "runner" not in data:
            raise ValueError(
                "No runner detected, please specify a runner in the config file."
            )

        runner = data["runner"]
        del data["runner"]

        options = []
        for kv in list(data.items()) + self.parse_manual_options(manual_options):
            options.append(self.parse_option(*kv))
        return f"{runner} {command} {' '.join(manual_arguments)} " + " ".join(options)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml")
    parser.add_argument("command")
    args, manual_inputs = parser.parse_known_args()

    try:
        manual_arguments, manual_options = " ".join(manual_inputs).split("--", 1)
        manual_arguments = manual_arguments.split()
        manual_options = ("--" + manual_options).split()
    except:
        manual_arguments = manual_inputs
        manual_options = []

    print(
        ConfigParser(args.yaml).parse_cmd(
            args.command,
            manual_arguments,
            manual_options,
        )
    )



class EMAMeter:
    def __init__(self, decay=0.98):
        self.decay = decay
        self.value = None

    def reset(self):
        self.value = None

    def __call__(self, value):
        if self.value is None:
            self.value = value
        else:
            self.value = self.decay * self.value + (1 - self.decay) * value
        return self.value


class Timer:
    def __init__(self, interval=1):
        self.interval = interval
        self.restart()

    def lap(self):
        return time.time() - self.start

    def timeup(self):
        return self.lap() > self.interval

    def restart(self):
        self.start = time.time()



import os
import tqdm
import time
import shutil
import argparse
import inspect
import numpy as np
import torch
import torch.nn as nn
import functools
import contextlib
import numbers
from pathlib import Path
from collections import defaultdict
from torch.utils.data import DataLoader
from functools import partial
from collections import defaultdict
from natsort import natsorted
from treelib import Tree

from torch.utils.tensorboard import SummaryWriter

import zouqi

from torchzq.parsing import boolean, flag, choices
from torchzq.saver import Saver
from torchzq.scheduler import Scheduler
from torchzq.events import Events
from torchzq.pbar import create_pbar


class Mode(str):
    logger = None
    events = None
    data_loader = None


class BaseRunner:
    model = None
    scheduler = None
    optimizer = None
    scaler = None
    saver = None

    modes = []

    def __init__(
        self,
        name: str = "Unnamed",
        batch_size: int = 32,
        nj: int = min(os.cpu_count(), 12),
        device: str = "cuda",
        strict_loading: boolean = True,
        ckpt_root: Path = "ckpt",
        logs_root: Path = "logs",
        quiet: flag = False,
        use_fp16: boolean = False,
        epoch: int = None,
        discard_optimizer_state_dict: boolean = False,
    ):
        self.update_args(locals(), "self")

    @property
    def mode(self):
        if len(self.modes) == 0:
            raise RuntimeError('No mode has been set, call "self.switch_mode()" first.')
        return self.modes[0]

    @property
    def data_loader(self):
        return self.mode.data_loader

    @property
    def dataset(self):
        return self.data_loader.dataset

    @property
    def logger(self):
        return self.mode.logger

    @property
    def name(self):
        return self.args.name

    @property
    def training(self):
        return self.mode == "train"

    @property
    def logs_dir(self):
        return self.args.logs_root / self.name / self.mode

    @property
    def ckpt_dir(self):
        return self.args.ckpt_root / self.name

    @property
    def events(self):
        return self.mode.events

    @property
    def autocast_if_use_fp16(self):
        return partial(torch.cuda.amp.autocast, enabled=self.args.use_fp16)

    @property
    def batch_size(self):
        return self.args.batch_size

    @property
    def Dataset(self):
        raise NotImplementedError

    @property
    def Optimizer(self):
        return torch.optim.Adam

    def update_args(self, payload, ignored=[]):
        if type(ignored) is str:
            ignored = [ignored]
        ignored += ["__class__"]
        for key in ignored:
            if key in payload:
                del payload[key]
        self.args = getattr(self, "args", argparse.Namespace())
        for k, v in payload.items():
            setattr(self.args, k, v)

    def autofeed(self, f, override={}, mapping={}):
        """Priority: 1. override, 2. parsed args 3. parameters' default"""
        assert hasattr(self, "args")
        payload = vars(self.args)

        def m(key):
            return mapping[key] if key in mapping else key

        signature = inspect.signature(getattr(f, "__init__", f))
        params = [p.name for p in signature.parameters.values()]

        kwargs = {k: payload[m(k)] for k in params if m(k) in payload}
        kwargs.update(override)

        return f(**kwargs)

    def create_scheduler(self):
        scheduler = Scheduler()
        if self.mode == "train":
            self.args.lr = scheduler.schedule(self.args.lr)
        return scheduler

    def create_dataset(self):
        raise NotImplementedError

    def create_data_loader(self, **kwargs):
        args = self.args
        dataset = self.create_dataset()
        data_loader = self.autofeed(
            DataLoader,
            override=dict(
                dataset=dataset,
                batch_size=self.batch_size,
                **kwargs,
            ),
            mapping=dict(
                num_workers="nj",
            ),
        )
        print("==> Dataset size:", len(dataset))
        return data_loader

    def create_model(self):
        raise NotImplementedError

    def create_optimizer(self, model):
        optimizer = self.autofeed(
            self.Optimizer,
            dict(params=model.parameters(), lr=1),
        )

        for state in optimizer.state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(self.args.device)

        def get_lr():
            return optimizer.param_groups[0]["lr"]

        def set_lr(lr):
            for g in optimizer.param_groups:
                g["lr"] = lr

        optimizer.get_lr = get_lr
        optimizer.set_lr = set_lr

        return optimizer

    def create_events(self):
        args = self.args
        names = [
            "iteration_started",
            "iteration_completed",
            "started",
            "completed",
        ]
        if self.training:
            names += [
                "epoch_started",
                "epoch_completed",
            ]
        events = Events(*names)

        if self.training:

            def save(epoch):
                if epoch % args.save_every == 0:
                    self.saver.save(self.model, self.optimizer, self.scaler)

            def validate(epoch):
                if epoch % args.validate_every == 0:
                    self.validate()
                    self.switch_mode("train")

            events.epoch_completed.append(save)
            events.epoch_completed.append(validate)

        return events

    def prepare_saver(self):
        if self.saver is None:
            self.saver = Saver(self.ckpt_dir, self.args.strict_loading)

    def prepare_events(self):
        if self.events is None:
            self.mode.events = self.create_events()

    def prepare_logger(self):
        if self.logger is None:
            self.mode.logger = SummaryWriter(self.logs_dir)

    def prepare_data_loader(self):
        # data_loader should be before the model
        if self.data_loader is None:
            self.mode.data_loader = self.create_data_loader(shuffle=self.training)

    def prepare_scheduler(self):
        # scheduler should be before the model
        if self.scheduler is None:
            self.scheduler = self.create_scheduler()

    def prepare_model(self):
        args = self.args
        if self.model is None:
            self.model = self.create_model()
            self.model.to(args.device)
            self.model.epoch = 0
            self.model.iteration = 0
            if self.training and not self.saver.empty and not args.continue_:
                print('Checkpoints exist and "--continue" not set, exited.')
                exit()
            self.saver.load(model=self.model, cache=True, epoch=args.epoch)
            self.scheduler.step(epoch=self.model.epoch, iteration=self.model.iteration)

    def prepare_optimizer(self):
        args = self.args
        if self.optimizer is None and self.training:
            self.optimizer = self.create_optimizer(self.model)
            if not args.discard_optimizer_state_dict:
                self.saver.load(optimizer=self.optimizer, epoch=args.epoch)

    def prepare_scaler(self):
        args = self.args
        if self.scaler is None and args.use_fp16:
            self.scaler = torch.cuda.amp.GradScaler()
            self.saver.load(scaler=self.scaler, epoch=args.epoch)

    def prepare_all(self):
        pipeline = lambda *levels: [f() for level in levels for f in level]
        pipeline(
            [self.prepare_saver, self.prepare_scheduler],
            [self.prepare_data_loader],
            [self.prepare_model],
            [self.prepare_optimizer, self.prepare_scaler],
            [self.prepare_events, self.prepare_logger],
        )

    def switch_mode(self, name):
        """
        Switch running mode.
        """
        # the first element in self.modes will be treated as the current mode
        if name in self.modes:
            i = self.modes.index(name)
            self.modes[0], self.modes[i] = self.modes[i], self.modes[0]
        else:
            self.modes.insert(0, Mode(name))
            self.prepare_all()
        if self.training:
            self.model.train()
        elif self.model is not None:
            self.model.eval()

    def step(self, batch):
        raise NotImplementedError

    @zouqi.command
    def train(
        self,
        lr: str = "1e-3",
        weight_decay: float = 0,
        max_epochs: int = 100,
        save_every: int = 1,
        validate_every: int = None,
        continue_: flag = False,
    ):
        if validate_every is None:
            validate_every = save_every

        self.update_args(locals(), "self")
        self.switch_mode("train")

        args = self.args
        model = self.model
        logger = self.logger

        self.events.started()
        while model.epoch < args.max_epochs:
            model.epoch += 1
            self.events.epoch_started(model.epoch)
            pbar = create_pbar(self.data_loader, args.quiet)
            pbar.set_description(f"Train: {model.epoch}/{args.max_epochs}")
            for batch in pbar:
                model.iteration += 1
                self.events.iteration_started(model.iteration)
                self.scheduler.step(model.epoch, model.iteration)
                try:
                    stats = self.step(batch)
                except RuntimeError as e:
                    if "out of memory" in str(e):
                        pbar.update_line(0, "OOM! Skip batch.")
                        for p in self.model.parameters():
                            if p.grad is not None:
                                del p.grad  # free some memory
                        torch.cuda.empty_cache()
                        continue
                    else:
                        raise e
                pbar.update_line(0, f"iteration: {model.iteration}")
                for l, (key, val) in enumerate(stats.items(), 1):
                    pbar.update_line(key, f"{key}: {val:.4g}")
                    logger.add_scalar(key, val, model.iteration)
                self.logger.flush()
                self.events.iteration_completed(model.iteration)
            pbar.close()
            self.events.epoch_completed(model.epoch)
        self.events.completed()

    @zouqi.command
    @torch.no_grad()
    def validate(self):
        self.update_args(locals(), "self")
        self.switch_mode("validate")

        args = self.args
        logger = self.logger
        model = self.model

        pbar = create_pbar(self.data_loader, args.quiet)
        pbar.set_description(f"Validate epoch: {model.epoch}")

        stats_list = defaultdict(list)
        self.events.started()
        for index, batch in enumerate(pbar):
            self.events.iteration_started(index)
            stats = self.step(batch)
            for l, (key, val) in enumerate(stats.items()):
                pbar.update_line(l, f"{key}: {val:.4g}")
                if isinstance(val, numbers.Number):
                    stats_list[key].append(val)
            self.events.iteration_completed(index)
        for key, val in stats_list.items():
            mean = sum(val) / len(val)
            self.logger.add_scalar(key, mean, model.epoch)
            print(f"Average {key}: {mean:.4g}.")
        self.logger.flush()
        self.events.completed()

    @staticmethod
    def try_rmtree(path):
        if path.exists():
            shutil.rmtree(path)
            print(str(path), "removed.")

    @zouqi.command
    def clear(self):
        self.ls()
        if input("Are you sure to clear? (y)\n").lower() == "y":
            self.try_rmtree(Path(self.args.ckpt_root, self.name))
            self.try_rmtree(Path(self.args.logs_root, self.name))
        else:
            print(f"Not cleared.")

    @zouqi.command
    def ls(self):
        self.print_tree(self.args.logs_root / self.name)
        self.print_tree(self.args.ckpt_root / self.name)

    @staticmethod
    def print_tree(root):
        tree = Tree()
        first = True

        for dirpath, _, files in os.walk(root):
            dirpath = Path(dirpath)

            if first:
                parent = None
            else:
                parent = dirpath.parent

            tree.create_node(
                tag=f"{dirpath if first else dirpath.name}/",
                identifier=dirpath,
                parent=parent,
            )

            first = False

            for f in natsorted(files):
                filepath = dirpath / f
                tree.create_node(
                    tag=f"{filepath.name} ({time.ctime(filepath.lstat().st_mtime)})",
                    identifier=filepath,
                    parent=dirpath,
                )

        if not first:
            tree.show()



import math


class _ScheduleFunction:
    def __init__(self, epochwise):
        self.epochwise = epochwise

    def step(self, epoch, iteration):
        if self.epochwise:
            self.n = epoch
        else:
            self.n = iteration

    def __deepcopy__(self, memo):
        print(
            "==> Warning: you are deepcopying a scheduler, which is forbidden. "
            "Shallow copy has been performed instead."
        )
        return self


class Cosine(_ScheduleFunction):
    def __init__(self, start, stop, epochwise):
        """https://www.tensorflow.org/api_docs/python/tf/compat/v1/train/cosine_decay
        Args:
            start: start
            stop: end
        """
        super().__init__(epochwise)
        self.start = start
        self.stop = stop
        assert 0 <= start < stop

    def __call__(self):
        pi = math.pi
        cos = math.cos
        start = self.start
        stop = self.stop

        if self.n < start:
            return 1
        if self.n > stop:
            return 0

        return max(0, 0.5 * (1 + cos(pi * (self.n - start) / (stop - start))))


class Exponential(_ScheduleFunction):
    def __init__(self, k, epochwise):
        super().__init__(epochwise)
        self.k = k
        assert k > 0

    def __call__(self):
        return math.exp(-self.k * self.n)


class Constant(_ScheduleFunction):
    def __init__(self, c):
        super().__init__(True)
        self.c = c

    def __call__(self):
        return self.c


class Lambda(_ScheduleFunction):
    def __init__(self, f, epochwise):
        super().__init__(epochwise)
        self.f = f

    def __call__(self):
        return self.f(self.n)


class Logistic(_ScheduleFunction):
    def __init__(self, k, start, upper, epochwise=True):
        super().__init__(epochwise)
        self.k = k
        self.start = start
        self.upper = upper
        assert k > 0

    def __call__(self):
        k = self.k
        start = self.start
        upper = self.upper
        return upper / (1 + math.exp(-k * (self.n - start)))


class Scheduler:
    def __init__(self):
        self._functions = []

    def schedule(self, x):
        x = eval(str(x))
        if not callable(x):
            x = Constant(x)
        self._functions.append(x)
        return x

    def step(self, epoch, iteration):
        for function in self._functions:
            function.step(epoch, iteration)



import torch
from pathlib import Path


def load_state_dict_safe(model, state_dict, strict):
    if strict:
        model.load_state_dict(state_dict, strict=True)
    else:
        model_state_dict = model.state_dict()
        provided = set(state_dict)
        required = set(model_state_dict)
        agreed = provided & required
        for k in list(agreed):
            if model_state_dict[k].shape != state_dict[k].shape:
                agreed.remove(k)
                provided.remove(k)
        state_dict = {k: state_dict[k] for k in agreed}
        print("Provided but not required keys: ")
        print(provided - required)
        print("Required but not provided keys: ")
        print(required - provided)
        model.load_state_dict(state_dict, strict=False)
    return model


class Saver:
    def __init__(self, root, strict=True):
        self.root = root
        self.strict = strict
        self.cache = {}

    @property
    def empty(self):
        return self.latest_epoch is None

    @property
    def latest_epoch(self):
        if not hasattr(self, "_latest_epoch") is None:
            paths = list(self.root.glob("*.pth"))
            if paths:
                self._latest_epoch = int(max(paths, key=lambda p: int(p.stem)).stem)
            else:
                self._latest_epoch = None
        return self._latest_epoch

    def read_state_dict(self, epoch, cache=False):
        path = self.root / f"{epoch}.pth"
        if path in self.cache:
            state_dict = self.cache[path]
        else:
            state_dict = torch.load(path, "cpu")
            if cache:
                self.cache[path] = state_dict
        return state_dict

    def load(
        self,
        model=None,
        optimizer=None,
        scaler=None,
        epoch=None,
        cache=False,
    ):
        epoch = self.latest_epoch if epoch is None else epoch

        if epoch is None:
            return

        state_dict = self.read_state_dict(epoch, cache)

        if model is not None:
            load_state_dict_safe(model, state_dict["model"], self.strict)
            model.epoch = epoch
            model.iteration = state_dict.get("iteration", 0)
            print(f"==> Model at epoch {epoch} loaded.")

        if optimizer is not None:
            try:
                optimizer.load_state_dict(state_dict["optimizer"])
                print(f"==> Optimizer at epoch {epoch} loaded.")
            except Exception as e:
                print(e)
                print("Warning: loading optimizer state dict failed.")

        if scaler is not None:
            try:
                scaler.load_state_dict(state_dict["scaler"])
                print(f"==> Scaler at epoch {epoch} loaded.")
            except Exception as e:
                print(e)
                print("Warning: loading scaler state dict failed.")

    def save(self, model, optimizer=None, scaler=None):
        self.root.mkdir(parents=True, exist_ok=True)
        state_dict = dict(
            iteration=model.iteration,
            epoch=model.epoch,
            model=model.state_dict(),
            optimizer=optimizer.state_dict() if optimizer else None,
            scaler=scaler.state_dict() if scaler else None,
        )
        path = self.root / f"{model.epoch}.pth"
        torch.save(state_dict, path)
        print(f"{path} saved.")



import tqdm
from collections import defaultdict

from torchzq.utils import Timer


def create_pbar(iterable, quiet):
    pbar = tqdm.tqdm(iterable, dynamic_ncols=True, disable=quiet)

    close = pbar.close

    create_line = lambda: tqdm.tqdm(bar_format="╰{postfix}")

    if quiet:
        items = {}
        line = create_line()

        timer = Timer(10)

        def update_line(i, s):
            items[i] = s
            if timer.timeup():
                line.set_postfix_str(", ".join(items.values()) + "\n")
                timer.restart()

        def close_all():
            close()
            line.close()

    else:
        lines = defaultdict(create_line)

        def update_line(i, s):
            lines[i].set_postfix_str(s)

        def close_all():
            close()
            for line in lines.values():
                line.close()

    pbar.update_line = update_line
    pbar.close = close_all

    return pbar
